from enum import Enum
from time import sleep as sync_sleep
from asyncio import sleep as async_sleep
from abc import ABC, abstractmethod
from threading import RLock
from dataclasses import dataclass
from typing import List, Dict, Awaitable, Optional, Union, Any
from collections.abc import Coroutine


from cantok.errors import CancellationError, SynchronousWaitingError


class CancelCause(Enum):
    CANCELLED = 1
    SUPERPOWER = 2
    NOT_CANCELLED = 3

@dataclass
class CancellationReport:
    cause: CancelCause
    from_token: 'AbstractToken'

class AngryAwaitable(Coroutine):  # type: ignore[type-arg]
    def __await__(self):  # type: ignore[no-untyped-def]
        raise SynchronousWaitingError('You cannot use the "await" keyword in the synchronous mode of the method. Add the "is_async" (bool) argument.')  # pragma: no cover
        yield self  # pragma: no cover

    def send(self, value: Any) -> None:
        raise SynchronousWaitingError('You cannot use the "await" keyword in the synchronous mode of the method. Add the "is_async" (bool) argument.')

    def throw(self, value: Any) -> Any:  # type: ignore[override]
        pass  # pragma: no cover

    def close(self) -> Any:
        pass  # pragma: no cover


class AbstractToken(ABC):
    exception = CancellationError
    rollback_if_nondirect_polling = False

    def __init__(self, *tokens: 'AbstractToken', cancelled: bool = False) -> None:
        self.tokens = tokens
        self._cancelled = cancelled
        self.lock = RLock()

    def __repr__(self) -> str:
        chunks = []
        superpower = self.text_representation_of_superpower()
        if superpower:
            chunks.append(superpower)
        other_tokens = ', '.join([repr(x) for x in self.tokens])
        if other_tokens:
            chunks.append(other_tokens)
        report = self.get_report(direct=False)
        if report.cause == CancelCause.NOT_CANCELLED:
            extra_kwargs = {}
        else:
            if report.from_token is self and report.cause == CancelCause.CANCELLED:
                extra_kwargs = {'cancelled': True}
            else:
                extra_kwargs = {}
        extra_kwargs.update(**(self.get_extra_kwargs()))
        text_representation_of_extra_kwargs = self.text_representation_of_kwargs(**extra_kwargs)
        if text_representation_of_extra_kwargs:
            chunks.append(text_representation_of_extra_kwargs)

        glued_chunks = ', '.join(chunks)
        return f'{type(self).__name__}({glued_chunks})'

    def __str__(self) -> str:
        cancelled_flag = 'cancelled' if self.is_cancelled(direct=False) else 'not cancelled'
        return f'<{type(self).__name__} ({cancelled_flag})>'

    def __add__(self, item: 'AbstractToken') -> 'AbstractToken':
        if not isinstance(item, AbstractToken):
            raise TypeError('Cancellation Token can only be combined with another Cancellation Token.')

        from cantok import SimpleToken

        return SimpleToken(self, item)

    def __bool__(self) -> bool:
        return self.keep_on()

    @property
    def cancelled(self) -> bool:
        return self.is_cancelled()

    @cancelled.setter
    def cancelled(self, new_value: bool) -> None:
        if new_value == True:
            self._cancelled = True
        else:
            if self._cancelled == True:
                raise ValueError('You cannot restore a cancelled token.')

    def keep_on(self) -> bool:
        return not self.is_cancelled()

    def is_cancelled(self, direct: bool = True) -> bool:
        return self.get_report(direct=direct).cause != CancelCause.NOT_CANCELLED

    def wait(self, step: Union[int, float] = 0.0001, timeout: Optional[Union[int, float]] = None, is_async: bool = False) -> Awaitable:  # type: ignore[type-arg]
        if step < 0:
            raise ValueError('The token polling iteration time cannot be less than zero.')
        if timeout is not None and timeout < 0:
            raise ValueError('The total timeout of waiting cannot be less than zero.')
        if timeout is not None and step > timeout:
            raise ValueError('The total timeout of waiting cannot be less than the time of one iteration of the token polling.')

        if timeout is None:
            from cantok import SimpleToken
            local_token: AbstractToken = SimpleToken()
        else:
            from cantok import TimeoutToken
            local_token = TimeoutToken(timeout)

        token = self + local_token

        async def async_wait() -> Awaitable:  # type: ignore[return, type-arg]
            while token:
                await async_sleep(step)

            local_token.check()

        def sync_wait() -> None:
            while token:
                sync_sleep(step)

            local_token.check()

        if is_async:
            return async_wait()

        else:
            sync_wait()
            return AngryAwaitable()


    def get_report(self, direct: bool = True) -> CancellationReport:
        if self._cancelled:
            return CancellationReport(
                cause=CancelCause.CANCELLED,
                from_token=self,
            )
        else:
            if self.check_superpower(direct):
                return CancellationReport(
                    cause=CancelCause.SUPERPOWER,
                    from_token=self,
                )

        for token in self.tokens:
            report = token.get_report(direct=False)
            if report.cause != CancelCause.NOT_CANCELLED:
                return report

        return CancellationReport(
            cause=CancelCause.NOT_CANCELLED,
            from_token=self,
        )

    def cancel(self) -> 'AbstractToken':
        self._cancelled = True
        return self

    @abstractmethod
    def superpower(self) -> bool:  # pragma: no cover
        pass

    def superpower_rollback(self, superpower_data: Dict[str, Any]) -> None:  # pragma: no cover
        pass

    def check_superpower(self, direct: bool) -> bool:
        if self.rollback_if_nondirect_polling and not direct:
            return self.check_superpower_with_rollback()
        return self.superpower()

    def check_superpower_with_rollback(self) -> bool:
        with self.lock:
            superpower_data = self.get_superpower_data()
            result = self.superpower()
            self.superpower_rollback(superpower_data)
            return result

    def get_superpower_data(self) -> Dict[str, Any]:  # pragma: no cover
        return {}

    @abstractmethod
    def text_representation_of_superpower(self) -> str:  # pragma: no cover
        pass

    def get_extra_kwargs(self) -> Dict[str, Any]:
        return {}

    def text_representation_of_extra_kwargs(self) -> str:
        return self.text_representation_of_kwargs(**(self.get_extra_kwargs()))

    def text_representation_of_kwargs(self, **kwargs: Any) -> str:
        pairs: List[str] = [f'{key}={repr(value)}' for key, value in kwargs.items()]
        return ', '.join(pairs)

    def check(self) -> None:
        with self.lock:
            report = self.get_report()

            if report.cause == CancelCause.CANCELLED:
                report.from_token.raise_cancelled_exception()

            elif report.cause == CancelCause.SUPERPOWER:
                report.from_token.raise_superpower_exception()

    def raise_cancelled_exception(self) -> None:
        raise CancellationError('The token has been cancelled.', self)

    def raise_superpower_exception(self) -> None:
        raise self.exception(self.get_superpower_exception_message(), self)

    @abstractmethod
    def get_superpower_exception_message(self) -> str:  # pragma: no cover
        return 'You have done the impossible to see this error.'



from cantok import AbstractToken
from cantok.errors import CancellationError


class SimpleToken(AbstractToken):
    exception = CancellationError

    def superpower(self) -> bool:
        return False

    def text_representation_of_superpower(self) -> str:
        return ''

    def get_superpower_exception_message(self) -> str:
        return 'The token has been cancelled.'  # pragma: no cover



from typing import Dict, Any

from cantok import AbstractToken
from cantok import ConditionToken
from cantok.errors import CounterCancellationError


class CounterToken(ConditionToken):
    exception = CounterCancellationError

    def __init__(self, counter: int, *tokens: AbstractToken, cancelled: bool = False, direct: bool = True):
        if counter < 0:
            raise ValueError('The counter must be greater than or equal to zero.')

        self.counter = counter
        self.initial_counter = counter
        self.direct = direct
        self.rollback_if_nondirect_polling = self.direct

        def function() -> bool:
            with self.lock:
                if not self.counter:
                    return True
                self.counter -= 1
                return False

        super().__init__(function, *tokens, cancelled=cancelled)

    def superpower_rollback(self, superpower_data: Dict[str, Any]) -> None:
        self.counter = superpower_data['counter']

    def text_representation_of_superpower(self) -> str:
        return str(self.counter)

    def get_extra_kwargs(self) -> Dict[str, Any]:
        return {
            'direct': self.direct,
        }

    def get_superpower_data(self) -> Dict[str, Any]:
        return {'counter': self.counter}

    def get_superpower_exception_message(self) -> str:
        return f'After {self.initial_counter} attempts, the counter was reset to zero.'



from typing import Any

class CancellationError(Exception):
    def __init__(self, message: str, token: Any):
        self.token = token
        super().__init__(message)

class ConditionCancellationError(CancellationError):
    pass

class CounterCancellationError(CancellationError):
    pass

class TimeoutCancellationError(CancellationError):
    pass

class SynchronousWaitingError(Exception):
    pass



from cantok.tokens.abstract_token import AbstractToken  # noqa: F401
from cantok.tokens.simple_token import SimpleToken  # noqa: F401
from cantok.tokens.condition_token import ConditionToken  # noqa: F401
from cantok.tokens.counter_token import CounterToken  # noqa: F401
from cantok.tokens.timeout_token import TimeoutToken

from cantok.errors import CancellationError, ConditionCancellationError, CounterCancellationError, TimeoutCancellationError  # noqa: F401


TimeOutToken = TimeoutToken



from time import monotonic_ns, perf_counter

from typing import Union, Callable, Dict, Any

from cantok import AbstractToken
from cantok import ConditionToken
from cantok.errors import TimeoutCancellationError


class TimeoutToken(ConditionToken):
    exception = TimeoutCancellationError

    def __init__(self, timeout: Union[int, float], *tokens: AbstractToken, cancelled: bool = False, monotonic: bool = False):
        if timeout < 0:
            raise ValueError('You cannot specify a timeout less than zero.')

        self.timeout = timeout
        self.monotonic = monotonic

        timer: Callable[[], Union[int, float]]
        if monotonic:
            timeout *= 1_000_000_000
            timer = monotonic_ns
        else:
            timer = perf_counter

        start_time: Union[int, float] = timer()
        def function() -> bool:
            return timer() >= (start_time + timeout)

        super().__init__(function, *tokens, cancelled=cancelled)

    def text_representation_of_superpower(self) -> str:
        return str(self.timeout)

    def get_extra_kwargs(self) -> Dict[str, Any]:
        return {
            'monotonic': self.monotonic,
        }

    def get_superpower_exception_message(self) -> str:
        return f'The timeout of {self.timeout} seconds has expired.'



from typing import Callable, Dict, Any
from contextlib import suppress

from cantok import AbstractToken
from cantok.errors import ConditionCancellationError


class ConditionToken(AbstractToken):
    exception = ConditionCancellationError

    def __init__(self, function: Callable[[], bool], *tokens: AbstractToken, cancelled: bool = False, suppress_exceptions: bool = True, default: bool = False):
        super().__init__(*tokens, cancelled=cancelled)
        self.function = function
        self.suppress_exceptions = suppress_exceptions
        self.default = default

    def superpower(self) -> bool:
        if not self.suppress_exceptions:
            return self.run_function()

        else:
            with suppress(Exception):
                return self.run_function()

        return self.default

    def run_function(self) -> bool:
        result = self.function()

        if not isinstance(result, bool):
            if not self.suppress_exceptions:
                raise TypeError(f'The condition function can only return a bool value. The passed function returned "{result}" ({type(result).__name__}).')
            else:
                return self.default

        return result

    def text_representation_of_superpower(self) -> str:
        return repr(self.function)

    def get_extra_kwargs(self) -> Dict[str, Any]:
        return {
            'suppress_exceptions': self.suppress_exceptions,
            'default': self.default,
        }

    def get_superpower_exception_message(self) -> str:
        return 'The condition is not met.'
